//using pgi compiler (openacc) on linux workstation with GPU (NVIDIA GeForce GTX 1080 Ti)
//following compile option is used (mt19937ar.cc and mt19937ar.h are for Mersenne twister)
//pgc++ topographical_guidance_Fig5_ridge.cc mt19937ar.cc -acc -Minfo -ta=tesla,cc60 -fast -Bstatic_pgi
//run: "a.out > data.dat"

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <math.h>
#include <iostream>
#include "mt19937ar.h"
#include "algorithm"
#include <algorithm>
#include <vector>
using namespace std;


#define pi 3.1415926536

//system size
#define L 256
#define Ly 256
#define Lz 150

//spatial and temporal resolutions
#define dx 0.2
#define dt 0.0002
#define cutoff 0.001

//parameters for membrane 
#define ep 1.2 //epsilon
#define eta 0.7
//#define eta 0.55 // use eta = 0.55 for Fig.5C 
#define Mem 5.0 //Mv
#define r0 6.0  //cell radius
#define beta_sig 100.0 
#define theta_sig 0.105 
#define tau 1.0

//parameters for protruding force
#define F0 2.6
#define K 0.05
#define Kd 0.04

//parameters for reaction-diffusion
#define Da 0.17
#define Di 0.13
#define ki1 0.05
#define ki2 0.5
#define at 1.6
#define gam2 10.0
#define xmax 50.0


//parameters for the initial condition
#define init_amp 5.0
#define init_rng 15


//noise parameters
#define seed 14
#define lam 0.001
#define sigma 1.5
#define d_noise 2.0


//parameters for substrate
#define Awall_rep 2.0
#define Awall_at 0.8

//total iteration is maxT*maxTT
int maxT=500;
#define maxTT 25000

//definition of grobal variables
double vol; //cell volume 
double area;//cell surface area
double intA;//total amount of A divided by cell surface area
double at_n; // at_n is normalized value of at and defined as at_n=at*4pi*R0*R0/area, where area is total surface area of the cell


//definition of field variables
int SIZE=L*Ly*Lz;
double* ph=(double*)std::malloc(sizeof(double)*(SIZE)); //phi
double* phn=(double*)std::malloc(sizeof(double)*(SIZE));
double* ps=(double*)std::malloc(sizeof(double)*(SIZE)); //psi
double* psn=(double*)std::malloc(sizeof(double)*(SIZE));
double* A=(double*)std::malloc(sizeof(double)*(SIZE));//concentration of 'A' molecule
double* An=(double*)std::malloc(sizeof(double)*(SIZE));
double* I=(double*)std::malloc(sizeof(double)*(SIZE));// concentration of 'I' molecule
double* In=(double*)std::malloc(sizeof(double)*(SIZE));
double* vamp=(double*)std::malloc(sizeof(double)*(SIZE));//the amplitude of the velocity
double* vx=(double*)std::malloc(sizeof(double)*(SIZE));// x component of the velocity vector
double* vy=(double*)std::malloc(sizeof(double)*(SIZE));// y component of the velocity vector
double* vz=(double*)std::malloc(sizeof(double)*(SIZE));// z component of the velocity vector
double* ch=(double*)std::malloc(sizeof(double)*(SIZE));//chi: field variable that represents the substrate


//function for caluclating psi from phi
double sigmoid(double x){
  return 1.0/(1.0+exp(-beta_sig*(x-theta_sig)));
}


int poisson(double Lam){
  Lam = exp(Lam) * genrand_real2();
  int k = 0;
   while (Lam > 1) {
     Lam *=genrand_real2();
     k++;
   }
   return k;
}


double Fpoly(double x1,double x2, double y1,double y2){
  int nh=3;
  double ret=0;
  if(x1>0){
    ret=(pow(x1,nh)/(pow(y1,nh)+pow(x1,nh)))*(pow(y2,nh)/(pow(y2,nh)+pow(x2,nh)));
  }  
  return ret;
}

//functions for spatial differentiation
double d1th_x(double *x, int i,int j,int k){
  return (x[(i+1)*Ly*Lz+j*Lz+k]-x[(i-1)*Ly*Lz+j*Lz+k])/(2.0*dx);
}
double d1th_y(double *x, int i,int j, int k){
  return (x[i*Ly*Lz+(j+1)*Lz+k]-x[i*Ly*Lz+(j-1)*Lz+k])/(2.0*dx);
}
double d1th_z(double *x, int i,int j, int k){
  return (x[i*Ly*Lz+j*Lz+k+1]-x[i*Ly*Lz+j*Lz+k-1])/(2.0*dx);
}

double d1th_x_3v(double *x,double *y,double *z, int i,int j, int k){
  return (x[(i+1)*Ly*Lz+j*Lz+k]*y[(i+1)*Ly*Lz+j*Lz+k]*z[(i+1)*Ly*Lz+j*Lz+k]-x[(i-1)*Ly*Lz+j*Lz+k]*y[(i-1)*Ly*Lz+j*Lz+k]*z[(i-1)*Ly*Lz+j*Lz+k])/(2.0*dx);
}
double d1th_y_3v(double *x,double *y,double *z, int i,int j,int k){
  return (x[i*Ly*Lz+(j+1)*Lz+k]*y[i*Ly*Lz+(j+1)*Lz+k]*z[i*Ly*Lz+(j+1)*Lz+k]-x[i*Ly*Lz+(j-1)*Lz+k]*y[i*Ly*Lz+(j-1)*Lz+k]*z[i*Ly*Lz+(j-1)*Lz+k])/(2.0*dx);
}
double d1th_z_3v(double *x,double *y,double *z, int i,int j,int k){
  return (x[i*Ly*Lz+j*Lz+k+1]*y[i*Ly*Lz+j*Lz+k+1]*z[i*Ly*Lz+j*Lz+k+1]-x[i*Ly*Lz+j*Lz+k-1]*y[i*Ly*Lz+j*Lz+k-1]*z[i*Ly*Lz+j*Lz+k-1])/(2.0*dx);
}

double d1th_abs(double *x, int i,int j,int k){
  double DX=d1th_x(x,i,j,k);
  double DY=d1th_y(x,i,j,k);
  double DZ=d1th_z(x,i,j,k);
  return sqrt(DX*DX+DY*DY+DZ*DZ);
}

double d2th(double *x, int i, int j,int k){
 return (x[(i+1)*Ly*Lz+j*Lz+k]+x[(i-1)*Ly*Lz+j*Lz+k]+x[(i)*Ly*Lz+(j+1)*Lz+k]+x[(i)*Ly*Lz+(j-1)*Lz+k]+x[i*Ly*Lz+j*Lz+k+1]+x[i*Ly*Lz+j*Lz+k-1]-6.0*x[i*Ly*Lz+j*Lz+k])/(dx*dx);
}

//function for calulating the phase-field equation
double func_vamp(double *AA,double *II,double *p, double *chi, int i, int j,int k, double VOL){
  double V0=  4.0*pi*r0*r0*r0/3.0;
  return (eta*(d2th(p,i,j,k)-16.0*p[i*Ly*Lz+j*Lz+k]*(1.0-p[i*Ly*Lz+j*Lz+k])*(1.0-2.0*p[i*Ly*Lz+j*Lz+k])/(ep*ep))+(-Mem*(VOL-V0)+F0*Fpoly(AA[i*Ly*Lz+j*Lz+k],II[i*Ly*Lz+j*Lz+k],K,Kd) )*d1th_abs(p,i,j,k)
	  -Awall_rep*chi[i*Lz*Ly+j*Lz+k]*chi[i*Lz*Ly+j*Lz+k]*p[i*Lz*Ly+j*Lz+k]+Awall_at*d1th_abs(chi,i,j,k)*d1th_abs(p,i,j,k)
	  )/tau;
}


//functions for calulating reaction-diffusion
double func_A(double *AA,double *II,double AT, double intb,int i, int j,int k){
  return  AA[i*Ly*Lz+j*Lz+k]*AA[i*Ly*Lz+j*Lz+k]*(AT-intb)/((1.0+II[i*Ly*Lz+j*Lz+k]))-AA[i*Ly*Lz+j*Lz+k];  
}
double func_I(double *BB,double *II, int i, int j, int k){
  return ki1*BB[i*Ly*Lz+j*Lz+k]*BB[i*Ly*Lz+j*Lz+k]-ki2*II[i*Ly*Lz+j*Lz+k];
}


//functions for boundary condition (periodic B.C. for x and y coordinate, the Dirichlet B.C. for z coordinate)
void BC_ph(double *x){

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int j=0;j<Ly;++j){
     x[i*Ly*Lz+j*Lz+0]=0;
     x[i*Ly*Lz+j*Lz+Lz-1]=0;
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2) 
 for(int j=0;j<Ly;++j){
   for(int k=0;k<Lz;++k){
     x[0*Ly*Lz+j*Lz+k]=x[(L-2)*Ly*Lz+j*Lz+k];
     x[(L-1)*Ly*Lz+j*Lz+k]=x[(1)*Ly*Lz+j*Lz+k];
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int k=0;k<Lz;++k){
     x[i*Ly*Lz+0*Lz+k]=x[i*Ly*Lz+(Ly-2)*Lz+k];
     x[i*Ly*Lz+(Ly-1)*Lz+k]=x[i*Ly*Lz+1*Lz+k];
   }
 }

}



void BC(double *x){
#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
  for(int i=0;i<L;++i){
    for(int j=0;j<Ly;++j){
      x[i*Ly*Lz+j*Lz+0]   =0;
      x[i*Ly*Lz+j*Lz+Lz-1]=0;
    }
  }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int j=0;j<Ly;++j){
   for(int k=0;k<Lz;++k){
     x[0*Ly*Lz+j*Lz+k]=x[(L-2)*Ly*Lz+j*Lz+k];
     x[(L-1)*Ly*Lz+j*Lz+k]=x[(1)*Ly*Lz+j*Lz+k];
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int k=0;k<Lz;++k){
     x[i*Ly*Lz+0*Lz+k]=x[i*Ly*Lz+(Ly-2)*Lz+k];
     x[i*Ly*Lz+(Ly-1)*Lz+k]=x[i*Ly*Lz+1*Lz+k];
   }
 }

}

//functions for update
void update_ph(double *x, double *xn){
#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
  for(int i=1;i<L-1;++i){
    for(int j=1;j<Ly-1;++j){
      for(int k=1;k<Lz-1;++k){
	x[i*Ly*Lz+j*Lz+k]=(xn[i*Ly*Lz+j*Lz+k]>=0)? xn[i*Ly*Lz+j*Lz+k]:0;
      }
    }
  }
  

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int j=0;j<Ly;++j){
     x[i*Ly*Lz+j*Lz+0]   =0;
     x[i*Ly*Lz+j*Lz+Lz-1]=0;
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2) 
 for(int j=0;j<Ly;++j){
   for(int k=0;k<Lz;++k){
     x[0*Ly*Lz+j*Lz+k]=x[(L-2)*Ly*Lz+j*Lz+k];
     x[(L-1)*Ly*Lz+j*Lz+k]=x[(1)*Ly*Lz+j*Lz+k];
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int k=0;k<Lz;++k){
     x[i*Ly*Lz+0*Lz+k]=x[i*Ly*Lz+(Ly-2)*Lz+k];
     x[i*Ly*Lz+(Ly-1)*Lz+k]=x[i*Ly*Lz+1*Lz+k];
   }
 }

  
}

void update(double *x, double *xn){

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
  for(int i=0;i<L;++i){
    for(int j=0;j<Ly;++j){
      for(int k=0;k<Lz;++k){
	x[i*Ly*Lz+j*Lz+k]=(xn[i*Ly*Lz+j*Lz+k]>=0)? xn[i*Ly*Lz+j*Lz+k]:0;
	if(xn[i*Ly*Lz+j*Lz+k]>xmax){x[i*Ly*Lz+j*Lz+k]=xmax;}
      }
    }
  }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int j=0;j<Ly;++j){
     x[i*Ly*Lz+j*Lz+0]   =0;
     x[i*Ly*Lz+j*Lz+Lz-1]=0;
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int j=0;j<Ly;++j){
   for(int k=0;k<Lz;++k){
     x[0*Ly*Lz+j*Lz+k]=x[(L-2)*Ly*Lz+j*Lz+k];
     x[(L-1)*Ly*Lz+j*Lz+k]=x[(1)*Ly*Lz+j*Lz+k];
   }
 }

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(2)
 for(int i=0;i<L;++i){
   for(int k=0;k<Lz;++k){
     x[i*Ly*Lz+0*Lz+k]=x[i*Ly*Lz+(Ly-2)*Lz+k];
     x[i*Ly*Lz+(Ly-1)*Lz+k]=x[i*Ly*Lz+1*Lz+k];
   }
 }

}


//function for setting chi (i.e., the substrate field)
void set_chi(){
  int z0 = 4*(ep/dx);
  for(int i=0;i<L;++i){
    for(int j=0;j<Ly;++j){
      for(int k=0;k<Lz;++k){
	ch[i*Ly*Lz+j*Lz+k]=0.5*(1.0+tanh((z0-k)/(0.5*(ep/dx)))) ;
      }
    }
  }

}


//initial condition
void set_init(){
#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)  
  for(int i=0;i<L;++i){
    for(int j=0;j<Ly;++j){
      for(int k=0;k<Lz;++k){
	double r1=sqrt((i-L/2)*(i-L/2)/(r0*r0)+(j-Ly/2)*(j-Ly/2)/(r0*r0)+(k-r0/dx)*(k-r0/dx)/(r0*r0))*dx;
	ph[i*Ly*Lz+j*Lz+k]=(1.0+tanh(2.0*(r0*(1.0-r1))/(ep)))/2.0;
	A[i*Ly*Lz+j*Lz+k]=0;
	I[i*Ly*Lz+j*Lz+k]=0;

      }
    }
  }

#pragma acc parallel loop
  for(int i=1;i<L-1;++i){
#pragma acc loop independent
    for(int j=1;j<Ly-1;++j){
#pragma acc loop independent
      for(int k=1;k<Lz-1;++k){
	ps[i*Ly*Lz+j*Lz+k]=sigmoid(ph[i*Ly*Lz+j*Lz+k]*(1.0-ph[i*Ly*Lz+j*Lz+k]));
      }
    }
  }

  BC(ps);

  vol=0;
  intA=0;
  area=0;
  set_chi();   
}

void paramet_logs(){
  /*
  time_t t = time(NULL);
  char filename[50];
  //sprintf(filename,"%s/log_code/%ld_%s",getenv("HOME"),t,__FILE__);
  sprintf(filename,"%s/log_code/%s_%ld",getenv("HOME"),__FILE__,t);
    FILE *fp,*fpw;
  fpw = fopen(filename, "w");
  fp = fopen(__FILE__, "r");
  char cha; 
  while( ( cha = fgetc(fp) ) != EOF ) {
    fprintf(fpw,"%c", cha);
  }
  fclose(fp);
  fclose(fpw);
  */
  
  cout<<" L Lz dx dt cutoff ep Mem r0 tau F0 gam2 beta_sig theta_sig eta K Kd Da Di ki1 ki2 at "<<endl;
  //  cout<<"#"<<filename<<" "<<L<<" "<<Lz<<" "<<dx<<" "<<dt<<" "<<cutoff<<" "<<ep<<" "<<Mem<<" "<<r0<<" "<<tau<<" "<<F0<<" "<<gam2<<" "<<beta_sig<<" "<<theta_sig<<" "<<eta<<" "<<K<<" "<<Kd<<" "<<Da<<" "<<Di<<" "<<ki1<<" "<<ki2<<" "<<at<<" "<<endl;
  cout<<"#"<<" "<<L<<" "<<Lz<<" "<<dx<<" "<<dt<<" "<<cutoff<<" "<<ep<<" "<<Mem<<" "<<r0<<" "<<tau<<" "<<F0<<" "<<gam2<<" "<<beta_sig<<" "<<theta_sig<<" "<<eta<<" "<<K<<" "<<Kd<<" "<<Da<<" "<<Di<<" "<<ki1<<" "<<ki2<<" "<<at<<" "<<endl;

}

//function for the initial patch nucleation
void set_initA(){
  for(int i=0;i<L;++i){
    for(int j=0;j<Ly;++j){
      for(int k=0;k<Lz;++k){
	if( sqrt((i-L/2)*(i-L/2)+(j-Ly/2)*(j-Ly/2))<=init_rng && ch[i*Ly*Lz+j*Lz+k]>cutoff && ps[i*Ly*Lz+j*Lz+k]>cutoff){
	  A[i*Ly*Lz+j*Lz+k]=init_amp*pow(genrand_real3(),5.0);
	}
      }
    }
  }
}

   
int main(){
  init_genrand(seed);
  paramet_logs();
  set_init();
  for(int T=0;T<maxT;++T){

      double t=1.0*T*maxTT*dt;
      
      //output data (only values for psi > 100*cutoff is recorded for compression of data)
      cout<<"#T="<<T<< " t="<<t<<" "<<vol<<" "<<area<<endl;
      
      for(int i=0;i<L;++i){
	for(int j=0;j<Ly;++j){
	  for(int k=0;k<Lz;++k){
	    if(ps[i*Ly*Lz+j*Lz+k]>100.0*cutoff ){cout<<i*Ly*Lz+j*Lz+k<<" "<<ph[i*Ly*Lz+j*Lz+k]<<" "<<ps[i*Ly*Lz+j*Lz+k]<<" "<<A[i*Ly*Lz+j*Lz+k]<<" "<<I[i*Ly*Lz+j*Lz+k]<<" "<<ch[i*Ly*Lz+j*Lz+k]<<endl;}
	  }
	}
      }
      

#pragma acc data create(vamp[0:L*Ly*Lz]),create(vx[0:L*Ly*Lz]),create(vy[0:L*Ly*Lz]),create(vz[0:L*Ly*Lz]),create(phn[0:L*Ly*Lz]),create(psn[0:L*Ly*Lz]),create(An[0:L*Ly*Lz]),create(In[0:L*Ly*Lz]),copy(ph[0:L*Ly*Lz]),copy(ps[0:L*Ly*Lz]),copy(A[0:L*Ly*Lz]),copy(I[0:L*Ly*Lz]),copyin(ch[0:L*Ly*Lz])
      for(int TT=0;TT<maxTT;++TT){

	
	vol=0;
	intA=0;
	area=0;

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
	for(int i=1;i<L-1;++i){
	  for(int j=1;j<Ly-1;++j){
	    for(int k=1;k<Lz-1;++k){
	      vol+=ph[i*Ly*Lz+j*Lz+k]*dx*dx*dx;
	      area+=ps[i*Ly*Lz+j*Lz+k]*dx*dx*dx/ep;
	      intA+=ps[i*Ly*Lz+j*Lz+k]*A[i*Ly*Lz+j*Lz+k]*dx*dx*dx;
	    }
	  }
	}
	
	double A0=4.0*pi*r0*r0;;
	intA=intA/(1.0*area);
	at_n=at*A0/(area);

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
	for(int i=1;i<L-1;++i){
	  for(int j=1;j<Ly-1;++j){
	    for(int k=1;k<Lz-1;++k){
	      phn[i*Ly*Lz+j*Lz+k]=ph[i*Ly*Lz+j*Lz+k]+dt*func_vamp(A,I,ph,ch,i,j,k,vol);
	      if(d1th_abs(ph,i,j,k)>cutoff){
		vamp[i*Ly*Lz+j*Lz+k]=func_vamp(A,I,ph,ch,i,j,k,vol)/d1th_abs(ph,i,j,k);
		vx[i*Ly*Lz+j*Lz+k]=-d1th_x(ph,i,j,k)/d1th_abs(ph,i,j,k);
		vy[i*Ly*Lz+j*Lz+k]=-d1th_y(ph,i,j,k)/d1th_abs(ph,i,j,k);
		vz[i*Ly*Lz+j*Lz+k]=-d1th_z(ph,i,j,k)/d1th_abs(ph,i,j,k);
	      }
	      else{
		vamp[i*Ly*Lz+j*Lz+k]=0;
	      }
	    }
	  }
	}
		
	BC(phn);BC(vamp);
	BC(vx);	BC(vy);	BC(vz);

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
	for(int i=1;i<L-1;++i){
	  for(int j=1;j<Ly-1;++j){
	    for(int k=1;k<Lz-1;++k){
	      psn[i*Ly*Lz+j*Lz+k]=sigmoid(phn[i*Ly*Lz+j*Lz+k]*(1.0-phn[i*Ly*Lz+j*Lz+k]));
	    }
	  }
	}

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
	for(int i=1;i<L-1;++i){
	  for(int j=1;j<Ly-1;++j){
	    for(int k=1;k<Lz-1;++k){
	      if(ps[i*Ly*Lz+j*Lz+k]>cutoff){
		An[i*Ly*Lz+j*Lz+k]=A[i*Ly*Lz+j*Lz+k]+dt*(-(d1th_x_3v(A,vamp,vx,i,j,k)+d1th_y_3v(A,vamp,vy,i,j,k)+d1th_z_3v(A,vamp,vz,i,j,k)) +beta_sig*(1.0-ps[i*Ly*Lz+j*Lz+k])*(1.0-2.0*ph[i*Ly*Lz+j*Lz+k])*(Da*(d1th_x(ph,i,j,k)*d1th_x(A,i,j,k)+d1th_y(ph,i,j,k)*d1th_y(A,i,j,k)+d1th_z(ph,i,j,k)*d1th_z(A,i,j,k)))+Da*d2th(A,i,j,k)+ func_A(A,I,at_n,intA,i,j,k));
		In[i*Ly*Lz+j*Lz+k]=I[i*Ly*Lz+j*Lz+k]+dt*(-(d1th_x_3v(I,vamp,vx,i,j,k)+d1th_y_3v(I,vamp,vy,i,j,k)+d1th_z_3v(I,vamp,vz,i,j,k)) +beta_sig*(1.0-ps[i*Ly*Lz+j*Lz+k])*(1.0-2.0*ph[i*Ly*Lz+j*Lz+k])*(Di*(d1th_x(ph,i,j,k)*d1th_x(I,i,j,k)+d1th_y(ph,i,j,k)*d1th_y(I,i,j,k)+d1th_z(ph,i,j,k)*d1th_z(I,i,j,k)))+Di*d2th(I,i,j,k)+ func_I(A,I,i,j,k));
	      }
	      else{
		An[i*Ly*Lz+j*Lz+k]=A[i*Ly*Lz+j*Lz+k]-dt*gam2*A[i*Ly*Lz+j*Lz+k];
		In[i*Ly*Lz+j*Lz+k]=I[i*Ly*Lz+j*Lz+k]-dt*gam2*I[i*Ly*Lz+j*Lz+k];
	      }

	    }
	  }
	}

	update(ph,phn);
	update(ps,psn);
	update(A,An);
	update(I,In);


	//noise
	if(T>15){
	int n=poisson(1.0*lam*L*Ly*dt*dx*dx);
	for(int nn=0;nn<n;++nn){
	  int x0 = int(1+genrand_int32()%(L-2)); 
	  int y0 = int(1+genrand_int32()%(L-2)); 
	  int z0 = 4*(ep/dx);
	  double amp=-sigma*log(genrand_real2());

#pragma acc parallel vector_length(256)
#pragma acc loop collapse(3)
	for(int i=1;i<L-1;++i){
	  for(int j=1;j<Ly-1;++j){
	    for(int k=1;k<Lz-1;++k){
	      if(fabs(x0-i)<=30 && fabs(y0-j)<=30 && fabs(z0-k)<=30 && ps[(i)*Ly*Lz+(j)*Lz+k]>cutoff && ch[(i)*Ly*Lz+(j)*Lz+k]>cutoff){
		A[(i)*Ly*Lz+(j)*Lz+k]+=amp*exp((-(x0-i)*(x0-i)-(y0-j)*(y0-j)-(z0-k)*(z0-k))/(2.0*d_noise*d_noise));
	      }
	    }
	  }
	}
	  

	}
	}

	

      }//close TT loop
      
      if(T==15){set_initA();}
      
    
     
  }//close T loop



  free(ph);
  free(phn);
  free(ps);
  free(psn);
  free(A);
  free(An);
  free(I);
  free(In);
  free(vamp);
  free(vx);
  free(vy);
  free(vz);

}



